package data;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Random;

import IO.DataReader;
import util.ArrayUtil;
import util.Permutation;
import util.PrintUtil;
import util.Ranking;

public class DataGenerator {
	
	private Random rand = new Random();
	
	public ArrayList<int[]> genMultiCoreUnsharpSet(int k, int N, int n, double p, int pk)
	{
		ArrayList<int[]> ret_A = new ArrayList<>();
		
		Permutation permutation = new Permutation(new Random());
		
		int[] perm;
		
		for(int i=0; i<k; i++)
		{
			perm = permutation.getRandomPermutation(n);
			
			for(int j=0; j<N/k; j++)
			{
				int[] ranking = new int[perm.length];
				
				for(int l=0; l<ranking.length; l++)
				{
					ranking[l] = perm[l];
				}
				
				ret_A.add(ranking);
			}
		}
			
		for(int i=0; i<pk; i++)
		{
			this.genIRDD(ret_A, p);
		}
		
		return ret_A;
		
	}
	
	public void runMultiData(int N)
	{
		String url = "res\\sushi_R.csv";
		
		boolean isOrdering = false;
		
		ArrayList<int[]> data = DataReader.readRankingData(url, ",", isOrdering);;		
		
		double p = 0.01;
		
		//Store Parameter
		StringBuffer sb = new StringBuffer();
		sb.append("# N: " + N + " p: " + p + " - sushi - \n");
		url = "C:\\Users\\client\\Desktop\\PHD\\MiningClosedRankPatterns\\data\\SynSushi_N" + N + "p" + p + "_R.txt";
		
		FileWriter fw;
		BufferedWriter bw;
		try {
			File file = new File(url);
			file.getParentFile().mkdirs();
			fw = new FileWriter(file,true);
			bw = new BufferedWriter(fw);
						
			bw.write(sb.toString());

			bw.close();
			fw.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		multiData(data, url, p, N);
		
	}
	
	public void runExtendRanking()
	{
		int N = 1;
		double p = 0.01;
		int numPairs = 1;
		
		for(int ex=1; ex<=10; ex++)
		{		
			String url = "C:\\Users\\client\\Desktop\\PHD\\MiningClosedRankPatterns\\data\\extended1\\SynSushi_N" + N + "p" + p + "ex" + (ex-1) + "_R.txt";
			
			boolean isOrdering = false;
			
			ArrayList<int[]> data = DataReader.readRankingData(url, ",", isOrdering);;		
	
			String desUrl = "C:\\Users\\client\\Desktop\\PHD\\MiningClosedRankPatterns\\data\\extended1\\SynSushi_N" + N + "p" + p + "ex" + ex + "_R.txt";
			int[][] extData = extendRankings(data, numPairs);
			
			FileWriter fw;
			BufferedWriter bw;
			try {
				File file = new File(desUrl);
				file.getParentFile().mkdirs();
				fw = new FileWriter(file,true);
				bw = new BufferedWriter(fw);
							
				for(int[] iA: extData)
				{
					for(int j=0; j<iA.length-1; j++)
					{
						bw.write(iA[j] + ",");
					}
					bw.write(iA[iA.length-1] + "\n");
				}
	
				bw.close();
				fw.close();
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		
	}
	
	public int[][] extendRankings(ArrayList<int[]> data, int np)
	{
		int n = data.get(0).length;
		int[][] extData = new int[data.size()][n+1];
		
		//select
		int[] pairsI = new int[np];
		for(int i=0; i<np; i++)
		{
			pairsI[i] = rand.nextInt(n)+1;
		}
	
		int[][] pairs = new int[np][2];
		for(int i=0; i<np; i++)
		{
			pairs[i] = Ranking.pairByIndex(pairsI[i], n);
		}
//System.out.println(pairs[0][0]);
//System.out.println(pairs[0][1]);
		int ia = 0;
		int ib = 0;
		int iE = 0; //index of the new element
		
		for(int r=0; r<data.size(); r++)
		{
			//select random pair
			int randPair = rand.nextInt(np);
			for(int i=0; i<data.get(r).length; i++)
			{				
				ia = data.get(r)[pairs[randPair][0]-1];
				ib = data.get(r)[pairs[randPair][1]-1];
			}
		
			if(ia < ib) iE = rand.nextInt(n+1-ib) + ib;
			else 		iE = rand.nextInt(n)+1;
//System.out.println(ia + " - " + ib);
//System.out.println(n+1-ib);
//System.out.println("iE: "+iE);			
			for(int i=0; i<n; i++)
			{
				if(data.get(r)[i] >= iE) 	extData[r][i] = data.get(r)[i] + 1;
				else						extData[r][i] = data.get(r)[i];		
			}
			
			extData[r][n] = iE;			
		}
		
		return extData;
	}
	
	public void multiData(ArrayList<int[]> data, String url, double p, int N)
	{
		ArrayList<int[]> copy_L;
		
		int[] ranking_A;
		
		try {
			FileWriter fw = new FileWriter(new File(url),true);
			BufferedWriter bw = new BufferedWriter(fw);

		
		for(int i=0; i<N; i++)
		{
			//copy data
			copy_L = new ArrayList<>();
			
			for(int[] iA: data)
			{
				ranking_A = new int[iA.length];
				
				for(int j=0; j<iA.length; j++)
				{
					ranking_A[j] = iA[j];
				}
				copy_L.add(ranking_A);
			}
System.out.println(PrintUtil.printArray(data.get(0), true));
System.out.println(PrintUtil.printArray(copy_L.get(0), true));
			
			//mutate Data
			genIRDD(copy_L, p);
System.out.println(PrintUtil.printArray(copy_L.get(0), true));
System.out.println("################");
		
			//store Data
			for(int[] iA: copy_L)
			{
				for(int j=0; j<iA.length-1; j++)
				{
					bw.write(iA[j] + ",");
				}
				bw.write(iA[iA.length-1] + "\n");
			}			
		}
		
		bw.close();
		fw.close();
		
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}
	
	public ArrayList<int[]> genNNaturalRankings(int N, int n)
	{
		ArrayList<int[]> ret_L = new ArrayList<>();
		
		int[] ranking;
		
		for(int i=0; i<N; i++)
		{
			ranking = new int[n];
			for(int j=0; j<n; j++)
			{
				ranking[j] = j+1;
			}
			ret_L.add(ranking);
		}
		
		return ret_L;
	}
	
	/**
	 * Randomises the given ranking by swapping every neighbored pair with probability p.
	 * @param rankings_L
	 * @param p
	 */
	public void genIRDD(ArrayList<int[]> rankings_L, double p)
	{
		int help;
		int[] ordering;
		
		for(int[] iA: rankings_L)
		{
			//transform into ordering first
			ordering = new int[iA.length];
			for(int i=0; i<iA.length; i++)
			{
				ordering[iA[i]-1] = i+1;
			}
		
			for(int i=0; i<ordering.length-1; i++)
			{
				if(rand.nextDouble() < p)
				{
					help = ordering[i];
					ordering[i] = ordering[i+1];
					ordering[i+1] = help;
				}
			}
			//transform back into ranking
			for(int i=0; i<iA.length; i++)
			{
				iA[ordering[i]-1] = i+1;
			}
		}
	}
	
	public ArrayList<int[]> stretchData(ArrayList<int[]> data, int N)
	{
		ArrayList<int[]> ret_L = new ArrayList<>();
		
		int dupFact = N/data.size();
		
		int[] help_A;
		
		for(int i=0; i<dupFact; i++)
		{
			for(int[] iA: data)
			{
				help_A = ArrayUtil.copy(iA);
				ret_L.add(help_A);
			}
		}
		
		int diff = N - ret_L.size();
		
		for(int i=0; i<diff; i++)
		{
			help_A = ArrayUtil.copy(data.get(i));
			ret_L.add(help_A);
		}
		
		return ret_L;
	}
}
