package miner.util;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;

import util.PrintUtil;

public class DataMaster {
	
	// ranking: permutation of objects
	// order: ordered objects

	public int[][] rankings;
	
//	private int[] objects; //every object which can be found in at least one ranking.
	
	//frequent Tuple
	public HashMap<Integer,HashMap<Integer,TransactionList>> 	frqTuple_M;
	public int[][]												followerMap;
	
	//#########################################################################
	// Constructor
	//#########################################################################
	
	public DataMaster(int[][] rankings)
	{
		this.rankings = rankings;
	}
	
	public DataMaster(ArrayList<int[]> rankings)
	{
		this.rankings = new int[rankings.size()][];
		rankings.toArray(this.rankings);
	}
	
	//#########################################################################
	// Business Methods
	//#########################################################################
	
	public void compFrq2Tuple(double t)
	{
		//frequent Tuple
		frqTuple_M = new HashMap<>();
		
		int numRanks = rankings[0].length;
		
		TWrapper[][] pairsCount = new TWrapper[numRanks][numRanks];
		
		// first iteration step -> initialization of TWrapper----------------
		for(int i=0; i<numRanks-1; i++)
		{
			for(int j=i+1; j<numRanks; j++)
			{
				pairsCount[i][j] = new TWrapper(new TransactionList(rankings.length),0);
				pairsCount[j][i] = new TWrapper(new TransactionList(rankings.length),0);
				
				if(rankings[0][i] < rankings[0][j] && rankings[0][i] > 0)
				{
					pairsCount[i][j].pairTail++;
					pairsCount[i][j].transaction.add(1);
				}
				else if(rankings[0][i] > rankings[0][j] && rankings[0][j] > 0)
				{
					pairsCount[j][i].pairTail++;
					pairsCount[j][i].transaction.add(1);
				}

			}
		}

//		for(int i=0; i<numRanks-1; i++)
//		{
//			for(int j=i+1; j<numRanks; j++)
//			{
//				System.out.println((i+1) + " : " + (j+1) + " | " + pairsCount[i][j].toString());
//				System.out.println((j+1) + " : " + (i+1) + " | " + pairsCount[j][i].toString());
//			}
//		}
		
		// rest iterations---------------------------------------------------
		for(int d=1; d<rankings.length; d++)
		{
			for(int i=0; i<numRanks-1; i++)
			{
				for(int j=i+1; j<numRanks; j++)
				{
					if(rankings[d][i] < rankings[d][j] && rankings[d][i] > 0)
					{
						pairsCount[i][j].pairTail++;
						pairsCount[i][j].transaction.add(d+1);
					}
					else if(rankings[d][i] > rankings[d][j] && rankings[d][j] > 0)
					{
						pairsCount[j][i].pairTail++;
						pairsCount[j][i].transaction.add(d+1);
					}
				}
			}
		}
		
		int help = 0;
		
		for(int i=0; i<numRanks-1; i++)
		{
			for(int j=i+1; j<numRanks; j++)
			{
				for(int k=0; k<2; k++)
				{
					if(pairsCount[i][j].pairTail >= rankings.length * t)
					{
						pairsCount[i][j].pairTail = j+1;
						
						if(frqTuple_M.containsKey(i+1))
						{
							frqTuple_M.get(i+1).put(pairsCount[i][j].pairTail, pairsCount[i][j].transaction);
						}
						else
						{
							frqTuple_M.put(i+1, new HashMap<Integer,TransactionList>());
							frqTuple_M.get(i+1).put(pairsCount[i][j].pairTail, pairsCount[i][j].transaction);
						}
					}
					
					help = i;
					i = j;
					j = help;
				}
			}
		}
	}
	
	public void compFrq2TupleV2(double t)
	{
		//frequent Tuple
		frqTuple_M = new HashMap<>();
		
		int numRanks = rankings[0].length;
		
		for(int i=0; i<numRanks; i++)
		{
			frqTuple_M.put(i+1,new HashMap<Integer,TransactionList>());
		}		
		
		TWrapper[][] pairsCount = new TWrapper[numRanks][numRanks];
		
		// first iteration step -> initialization of TWrapper----------------
		for(int i=0; i<numRanks-1; i++)
		{
			for(int j=i+1; j<numRanks; j++)
			{
				pairsCount[i][j] = new TWrapper(new TransactionList(rankings.length),0);
				pairsCount[j][i] = new TWrapper(new TransactionList(rankings.length),0);
				
				if(rankings[0][i] < rankings[0][j] && rankings[0][i] > 0)
				{
					pairsCount[i][j].pairTail++;
					pairsCount[i][j].transaction.add(1);
				}
				else if(rankings[0][i] > rankings[0][j] && rankings[0][j] > 0)
				{
					pairsCount[j][i].pairTail++;
					pairsCount[j][i].transaction.add(1);
				}

			}
		}
		
		// rest iterations---------------------------------------------------
		for(int d=1; d<rankings.length; d++)
		{
			for(int i=0; i<numRanks-1; i++)
			{
				for(int j=i+1; j<numRanks; j++)
				{
					if(rankings[d][i] < rankings[d][j] && rankings[d][i] > 0)
					{
						pairsCount[i][j].pairTail++;
						pairsCount[i][j].transaction.add(d+1);
					}
					else if(rankings[d][i] > rankings[d][j] && rankings[d][j] > 0)
					{
						pairsCount[j][i].pairTail++;
						pairsCount[j][i].transaction.add(d+1);
					}
				}
			}
		}
		
		int help = 0;
		
		for(int i=0; i<numRanks-1; i++)
		{
			for(int j=i+1; j<numRanks; j++)
			{
				for(int k=0; k<2; k++)
				{
					if(pairsCount[i][j].pairTail >= rankings.length * t)
					{
						pairsCount[i][j].pairTail = j+1;

						frqTuple_M.get(i+1).put(pairsCount[i][j].pairTail, pairsCount[i][j].transaction);		
					}
					
					help = i;
					i = j;
					j = help;
				}
			}
		}
	}
	
	public ArrayList<PreferenceNode> iMatrixToTree(int[][] iMatrix, ObjectTuple firstTA)
	{
		//copy iMatrix to keep the outer unchanged
		int[][] iMc = new int[iMatrix.length][iMatrix[0].length];
		for(int i=0; i<iMc.length; i++)
		{
			for(int j=0; j<iMc[i].length; j++)
			{
				iMc[i][j] = iMatrix[i][j];
			}
		}
		
		ArrayList<PreferenceNode> prefWood_L = new ArrayList<>();		
		ArrayList<PreferenceNode> prefGraph_L = new ArrayList<>();
		
		for(int i=0; i<firstTA.size(); i++)
		{
			prefGraph_L.add(new PreferenceNode(firstTA.get(i)));
		}
		
		int p = 0;
		int[] ancestors = new int[iMatrix.length];
		
		for(int i=0; i<prefGraph_L.size(); i++)
		{
			if(!prefGraph_L.get(i).getFlag())
			{
				prefWood_L.add(prefGraph_L.get(i));
				
				p = i;
				ancestors[0] = p;
				
				recursiveConnect(p,1,ancestors,iMc,prefGraph_L);
			}
		}	
		return prefWood_L;
	} 
	
	private void recursiveConnect(int p, int numA, int[] ancestors, int[][] iMatrix, ArrayList<PreferenceNode> prefGraph)
	{
		for(int i=p + 1; i<iMatrix.length; i++)
		{		
			if(iMatrix[i][p] == 0)
			{
				if(!prefGraph.get(p).getFlag()) prefGraph.get(p).addNode(prefGraph.get(i));
								
				for(int j=0; j<numA-1; j++)
				{
					iMatrix[i][ancestors[j]] = 1;
				}
				
				ancestors[numA] = i;
		
				recursiveConnect(i, numA + 1, ancestors, iMatrix, prefGraph);
			}
		}
		prefGraph.get(p).setFlag(true);

	}
	
	public int[][] intersectionMatrix(TransactionList tl)
	{
		// Transaction indices
		ArrayList<Integer> trans_L = tl.getTransactions();
		
		// get first transaction and convert it into an ObjectTuple
		ObjectTuple firstTA = new ObjectTuple(rankings[trans_L.get(0)-1]);
		
		int[][] ret_AA = new int[firstTA.size()][firstTA.size()];
		
		for(int o=1; o<firstTA.size(); o++)
		{
			for(int p=(o-1); p >= 0; p--)
			{
				if((this.frqTuple_M.get(firstTA.get(p)) == null)
				||(this.frqTuple_M.get(firstTA.get(p)).get(firstTA.get(o)) == null)
				||(!this.frqTuple_M.get(firstTA.get(p)).get(firstTA.get(o)).contains(tl)))
				{
					ret_AA[o][p] = 1;
				}
			}
		}		
		return ret_AA;
	}
	
	public int[][] intersectionMatrix(TransactionList tl, ArrayList<int[]> leftPairs)
	{
		// Transaction indices
		ArrayList<Integer> trans_L = tl.getTransactions();
		
		// get first transaction and convert it into an ObjectTuple
		ObjectTuple firstTA = new ObjectTuple(rankings[trans_L.get(0)-1]);
		
		int[][] ret_AA = new int[firstTA.size()][firstTA.size()];
		
		for(int o=1; o<firstTA.size(); o++)
		{
			for(int p=(o-1); p >= 0; p--)
			{
				if((this.frqTuple_M.get(firstTA.get(p)) == null)
				||(this.frqTuple_M.get(firstTA.get(p)).get(firstTA.get(o)) == null)
				||(!this.frqTuple_M.get(firstTA.get(p)).get(firstTA.get(o)).contains(tl)))
				{
					ret_AA[o][p] = 1;
					leftPairs.add(new int[]{p,o});
				}					
			}
		}		
		return ret_AA;
	}
	
	public void intersectionMatrix(TransactionList tl, int[][] iMatrix, ArrayList<int[]> leftPairs)
	{
		ArrayList<int[]> removeL = new ArrayList<>();
		for(int[] pairs: leftPairs)
		{
			if((this.frqTuple_M.get(pairs[0]) == null)
			||(this.frqTuple_M.get(pairs[0]).get(pairs[1]) == null)
			||(!this.frqTuple_M.get(pairs[0]).get(pairs[1]).contains(tl)))
			{
				iMatrix[pairs[1]][pairs[0]] = 1;
			}
			else
			{
				iMatrix[pairs[1]][pairs[0]] = 0;
				removeL.add(pairs);
			}
		}
		removeL.removeAll(removeL);
	}
	
//	public ArrayList<PreferenceNode> compHClosure(ObjectTuple ot)
//	{
//		int[][] iMatrix = intersectionMatrix(ot.getTransactionList());
//
//		// get first transaction and convert it into an ObjectTuple
//		ObjectTuple firstTA = new ObjectTuple(rankings[ot.getTransactionList().getTransactions().get(0)-1]);
//		
//		ArrayList<PreferenceNode> hClosure = this.iMatrixToTree(iMatrix, firstTA);
//		
//		return hClosure;
//	}
//	
//	public int[][] compHClosure(ObjectTuple ot, ArrayList<PreferenceNode> hClosure, ArrayList<int[]> leftPairs)
//	{
//		int[][] iMatrix = intersectionMatrix(ot.getTransactionList(), leftPairs);
//		// get first transaction and convert it into an ObjectTuple
//		ObjectTuple firstTA = new ObjectTuple(rankings[ot.getTransactionList().getTransactions().get(0)-1]);
//		
//		hClosure.addAll(this.iMatrixToTree(iMatrix, firstTA));
//		
//		return iMatrix;
//	}
//	
//	public void compHClosure(ObjectTuple ot, ArrayList<PreferenceNode> hClosure, int[][] iMatrix, ArrayList<int[]> leftPairs)
//	{
//		intersectionMatrix(ot.getTransactionList(),iMatrix, leftPairs);
//		// get first transaction and convert it into an ObjectTuple
//		ObjectTuple firstTA = new ObjectTuple(rankings[ot.getTransactionList().getTransactions().get(0)-1]);
//		
//		hClosure.addAll(this.iMatrixToTree(iMatrix, firstTA));
//	}
	
	public boolean isPrefix(ArrayList<PreferenceNode> tree, ObjectTuple ot)
	{
		PreferenceNode treeBase = new PreferenceNode(-1);
		for(PreferenceNode pn: tree)
		{
			treeBase.addNode(pn);
		}
		
		return recursivePrefSearch(treeBase,ot,0);
	}
	
	private boolean recursivePrefSearch(PreferenceNode pn, ObjectTuple ot, int index)
	{
		if(ot.size() <= index) return true;
			
		for(int i=0; i<pn.numChildren(); i++)
		{
			if(pn.getNode(i).getValue() == ot.get(index))
			{
				return recursivePrefSearch(pn.getNode(i), ot, index+1);
			}
		}
		return false;
	}
	
	public PreferenceNode isPrefixX2(ArrayList<PreferenceNode> tree, ObjectTuple ot)
	{
		PreferenceNode treeBase = new PreferenceNode(-1);
		for(PreferenceNode pn: tree)
		{
			treeBase.addNode(pn);
		}
		
		return recursivePrefSearchX2(treeBase,ot,0);
	}
	
	public PreferenceNode recursivePrefSearchX2(PreferenceNode pn, ObjectTuple ot, int index)
	{
		if(ot.size() <= index) return pn;
		
		for(int i=0; i<pn.numChildren(); i++)
		{
			if(pn.getNode(i).getValue() == ot.get(index))
			{
				return recursivePrefSearchX2(pn.getNode(i), ot, index+1);
			}
		}
		return null;
	}
	
	public int size()
	{
		return this.rankings.length;
	}
	
	public int numRanks()
	{
		return rankings[0].length;
	}
	
	public HashMap<Integer,HashMap<Integer,TransactionList>> getFrequentTuple()
	{
		return this.frqTuple_M;
	}
	
	//#########################################################################
	// static Methods
	//#########################################################################
	
	public static ArrayList<ArrayList<ObjectTuple>> extractClosed(ArrayList<ArrayList<ObjectTuple>> frq_LL)
	{
		ArrayList<ArrayList<ObjectTuple>> closed_LL = new ArrayList<>();
		
		return null;
	}
	
}
