package miner.util;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;

import util.PrintUtil;

public class DataMasterD2 {
	
	// ranking: permutation of objects
	// order: ordered objects

	public int[][] rankings;
	public PairVector[] rankingsPV;
	
//	private int[] objects; //every object which can be found in at least one ranking.
	
	//frequent Tuple
	public HashMap<Integer,HashMap<Integer,TransactionList>> 	frqTuple_M;
	public int[][]												followerMap;
	
	int[][] iMcL;
	int[] ancestorsL;
	int[] firstTA_A;
	int[] help;
	int 	c;
	PreferenceNode treeBaseL;
	PreferenceNode[][] prefGraph_L;
	
	//#########################################################################
	// Constructor
	//#########################################################################
	
	public DataMasterD2(int[][] rankings)
	{
		this.rankings = rankings;
		
		int n = rankings[0].length;
		
		iMcL = new int[n][n];
		ancestorsL = new int[n];
		firstTA_A = new int[n];
		help = new int[n];
		treeBaseL = new PreferenceNode(-1);
		prefGraph_L = new PreferenceNode[n][n*(n-1)/2];
		
		initPrefGraph(n);
		
//		rankingsPV = new PairVector[rankings.length];
//		for(int i=0; i<rankings.length; i++)
//		{
//			rankingsPV[i] = new PairVector(rankings[i]);
//		}
	}
	
	public DataMasterD2(ArrayList<int[]> rankings)
	{
		this.rankings = new int[rankings.size()][];
		rankings.toArray(this.rankings);
		
		int n = rankings.get(0).length;
		
		iMcL = new int[n][n];
		ancestorsL = new int[n];
		firstTA_A = new int[n];
		help = new int[n];
		treeBaseL = new PreferenceNode(-1);
		prefGraph_L = new PreferenceNode[n][n*(n-1)/2];
		
		initPrefGraph(n);
		
//		rankingsPV = new PairVector[this.rankings.length];
//		for(int i=0; i<this.rankings.length; i++)
//		{
//			rankingsPV[i] = new PairVector(this.rankings[i]);
//		}
	}
	
	public void initPrefGraph(int n)
	{
		for(int i=0; i<n; i++)
		for(int j=0; j<n*(n-1)/2; j++)
			prefGraph_L[i][j] = new PreferenceNode(-1);
	
	}
	
	//#########################################################################
	// Business Methods
	//#########################################################################
	
	public void compFrq2Tuple(double t)
	{
		//frequent Tuple
		frqTuple_M = new HashMap<>();
		
		int numRanks = rankings[0].length;
		
		TWrapper[][] pairsCount = new TWrapper[numRanks][numRanks];
		
		// first iteration step -> initialization of TWrapper----------------
		for(int i=0; i<numRanks-1; i++)
		{
			for(int j=i+1; j<numRanks; j++)
			{
				pairsCount[i][j] = new TWrapper(new TransactionList(rankings.length),0);
				pairsCount[j][i] = new TWrapper(new TransactionList(rankings.length),0);
				
				if(rankings[0][i] < rankings[0][j] && rankings[0][i] > 0)
				{
					pairsCount[i][j].pairTail++;
					pairsCount[i][j].transaction.add(1);
				}
				else if(rankings[0][i] > rankings[0][j] && rankings[0][j] > 0)
				{
					pairsCount[j][i].pairTail++;
					pairsCount[j][i].transaction.add(1);
				}

			}
		}
		
		// rest iterations---------------------------------------------------
		for(int d=1; d<rankings.length; d++)
		{
			for(int i=0; i<numRanks-1; i++)
			{
				for(int j=i+1; j<numRanks; j++)
				{
					if(rankings[d][i] < rankings[d][j] && rankings[d][i] > 0)
					{
						pairsCount[i][j].pairTail++;
						pairsCount[i][j].transaction.add(d+1);
					}
					else if(rankings[d][i] > rankings[d][j] && rankings[d][j] > 0)
					{
						pairsCount[j][i].pairTail++;
						pairsCount[j][i].transaction.add(d+1);
					}
				}
			}
		}
		
		int help = 0;
		
		for(int i=0; i<numRanks-1; i++)
		{
			for(int j=i+1; j<numRanks; j++)
			{
				for(int k=0; k<2; k++)
				{
					if(pairsCount[i][j].pairTail >= rankings.length * t)
					{
						pairsCount[i][j].pairTail = j+1;
						
						if(frqTuple_M.containsKey(i+1))
						{
							frqTuple_M.get(i+1).put(pairsCount[i][j].pairTail, pairsCount[i][j].transaction);
						}
						else
						{
							frqTuple_M.put(i+1, new HashMap<Integer,TransactionList>());
							frqTuple_M.get(i+1).put(pairsCount[i][j].pairTail, pairsCount[i][j].transaction);
						}
					}
					
					help = i;
					i = j;
					j = help;
				}
			}
		}
	}
	
	public ArrayList<PreferenceNode> iMatrixToTree(int[][] iMatrix, ObjectTuple firstTA)
	{
		//copy iMatrix to keep the outer unchanged
		int[][] iMc = new int[iMatrix.length][iMatrix[0].length]; //hiere ine globale im DataMaster erstellen.
		for(int i=0; i<iMc.length; i++)
		{
			System.arraycopy(iMatrix[i], 0, iMc[i], 0, iMc[i].length);
		}
		
		ArrayList<PreferenceNode> prefWood_L = new ArrayList<>();		
		ArrayList<PreferenceNode> prefGraph_L = new ArrayList<>();
		
		for(int i=0; i<firstTA.size(); i++)
		{
			prefGraph_L.add(new PreferenceNode(firstTA.get(i)));
		}
		
		int p = 0;
		int[] ancestors = new int[iMatrix.length];
		
		for(int i=0; i<prefGraph_L.size(); i++)
		{
			if(iMc[firstTA.get(i)-1][firstTA.get(i)-1] == 0)
			{
				prefWood_L.add(prefGraph_L.get(i));
				
				p = i;
				ancestors[0] = p;
				
				recursiveConnect(p, 1, ancestors, iMc, prefGraph_L, firstTA);
			}
		}	
		return prefWood_L;
	}
	
	public void iMatrixToTree(ArrayList<PreferenceNode> hClosure, int[][] iMatrix, int length, int k)
	{
		//copy iMatrix to keep the outer unchanged
		for(int i=0; i<iMcL.length; i++)
		{
			System.arraycopy(iMatrix[i], 0, iMcL[i], 0, iMcL[i].length);
		}
		
		for(int i=0; i<length; i++)
		{
			prefGraph_L[k][i].clearUp(firstTA_A[i]);
		}
		
		int p = 0;
		
		for(int i=0; i<iMatrix.length; i++)
			ancestorsL[i] = 0;
		
		for(int i=0; i<length; i++)
		{
			if(iMcL[firstTA_A[i]-1][firstTA_A[i]-1] == 0)
			{
				hClosure.add(prefGraph_L[k][i]);
				
				p = i;
				ancestorsL[0] = p;
				
				recursiveConnect2(p, 1, iMcL, length, k);
			}
		}	
	}
	
	private void recursiveConnect(int p, int numA, int[] ancestors, int[][] iMatrix, ArrayList<PreferenceNode> prefGraph, ObjectTuple firstTA)
	{
		for(int i=p + 1; i<prefGraph.size(); i++)
		{		
			if(iMatrix[firstTA.get(p)-1][firstTA.get(i)-1] == 1)
			{
				if(iMatrix[firstTA.get(p)-1][firstTA.get(p)-1] == 0) prefGraph.get(p).addNode(prefGraph.get(i));
								
				for(int j=0; j<numA-1; j++)
				{
					iMatrix[firstTA.get(ancestors[j])-1][firstTA.get(i)-1] = 0;
				}				
				ancestors[numA] = i;		
				recursiveConnect(i, numA + 1, ancestors, iMatrix, prefGraph, firstTA);
			}
		}
		iMatrix[firstTA.get(p)-1][firstTA.get(p)-1] = 1;
	}
	
	private void recursiveConnect2(int p, int numA, int[][] iMatrix, int length, int k)
	{
		for(int i=p + 1; i<length; i++)
		{		
			if(iMatrix[firstTA_A[p]-1][firstTA_A[i]-1] == 1)
			{
				if(iMatrix[firstTA_A[p]-1][firstTA_A[p]-1] == 0) prefGraph_L[k][p].addNode(prefGraph_L[k][i]);
								
				for(int j=0; j<numA-1; j++)
				{
					iMatrix[firstTA_A[ancestorsL[j]]-1][firstTA_A[i]-1] = 0;
				}				
				ancestorsL[numA] = i;		
				recursiveConnect2(i, numA + 1, iMatrix, length, k);
			}
		}
		iMatrix[firstTA_A[p]-1][firstTA_A[p]-1] = 1;
	}
	
	public void intersectionMatrix(TransactionList tl, int[][] iMatrix, ArrayList<int[]> leftPairs, ArrayList<int[]> removeL)
	{		
		for(int[] pairs: leftPairs)
		{
			if(this.frqTuple_M.get(pairs[0]).get(pairs[1]).contains(tl))
			{
				iMatrix[pairs[0]-1][pairs[1]-1] = 1;
				removeL.add(pairs);
			}
		}
	}
	
	public void compHClosure(TransactionList tl, ArrayList<PreferenceNode> hClosure, int[][] iMatrix, ArrayList<int[]> leftPairs, ArrayList<int[]> removeL)
	{
		intersectionMatrix(tl,iMatrix, leftPairs, removeL);
		// get first transaction and convert it into an ObjectTuple
		ObjectTuple firstTA = new ObjectTuple(rankings[tl.getFirst()-1]);
		
		hClosure.addAll(this.iMatrixToTree(iMatrix, firstTA));
	}
	
	public void compHClosure2(TransactionList tl, ArrayList<PreferenceNode> hClosure, int[][] iMatrix, ArrayList<int[]> leftPairs, ArrayList<int[]> removeL, int k)
	{
		intersectionMatrix(tl,iMatrix, leftPairs, removeL);
		// get first transaction and convert it into an ObjectTuple	
		int length = rankingToOrdering(firstTA_A, rankings[tl.getFirst()-1]);
		
		this.iMatrixToTree(hClosure, iMatrix, length, k);
	}
	
	public PreferenceNode isPrefixX2(ArrayList<PreferenceNode> tree, ObjectTuple ot)
	{
		PreferenceNode treeBase = new PreferenceNode(-1);
		for(PreferenceNode pn: tree)
		{
			treeBase.addNode(pn);
		}
		
		return recursivePrefSearchX2(treeBase,ot,0);
	}
	
	public PreferenceNode recursivePrefSearchX2(PreferenceNode pn, ObjectTuple ot, int index)
	{
		if(ot.size() <= index) return pn;
		
		for(int i=0; i<pn.numChildren(); i++)
		{
			if(pn.getNode(i).getValue() == ot.get(index))
			{
				return recursivePrefSearchX2(pn.getNode(i), ot, index+1);
			}
		}
		return null;
	}
	
	public PreferenceNode isPrefixX2_2(ArrayList<PreferenceNode> tree, ObjectTuple ot)
	{
		treeBaseL.clear();
		for(PreferenceNode pn: tree)
		{
			treeBaseL.addNode(pn);
		}
		
		return recursivePrefSearchX2(treeBaseL,ot,0);
	}

//	public void recursivePrefSearchX2(PreferenceNode postClosure, PreferenceNode pn, ObjectTuple ot, int index)
//	{
//		if(ot.size() > index)		
//		for(int i=0; i<pn.numChildren(); i++)
//		{
//			if(pn.getNode(i).getValue() == ot.get(index))
//			{
//				return recursivePrefSearchX2(pn.getNode(i), ot, index+1);
//			}
//		}
//	}
	
	public int size()
	{
		return this.rankings.length;
	}
	
	public int numRanks()
	{
		return rankings[0].length;
	}
	
	public HashMap<Integer,HashMap<Integer,TransactionList>> getFrequentTuple()
	{
		return this.frqTuple_M;
	}
	
	public int rankingToOrdering(int[] ordering, int[] ranking)
	{			
		c = 1;
		for(int i=0; i<ordering.length; i++)
		{
			if(ranking[i] > 0) help[ranking[i]-1] = i+1;
			else
			{
				help[ranking.length - c] = 0;
				c++;
			}
		}
		
		int count = 0;
		for(int i=0; i<help.length; i++)
		{
			if(help[i] > 0)
			{
				ordering[count] = help[i];
				count++;
			}
			else break;
		}
		
		return count;
	}
	
	//#########################################################################
	// static Methods
	//#########################################################################
	
	public static ArrayList<ArrayList<ObjectTuple>> extractClosed(ArrayList<ArrayList<ObjectTuple>> frq_LL)
	{
		ArrayList<ArrayList<ObjectTuple>> closed_LL = new ArrayList<>();
		
		return null;
	}
	
}
