package miner.util;

import java.util.ArrayList;

public class TransactionList {	
	
	public int[] list;
	
	private int hashValue = 0;
	
	//#########################################################################
	// Constructor
	//#########################################################################
	
	public TransactionList(int[] tset)
	{	
		list = TransactionsToBits(tset, tset[tset.length-1]);
	}
	
	public TransactionList(int bN)
	{
		list = new int[(int)Math.ceil(bN/(double)31)];
	}
	
	//#########################################################################
	// static Methods
	//#########################################################################
	
	public static int[] TransactionsToBits(int[] tset, int bN)
	{
		
		int[] ret = new int[(int)Math.ceil(bN/(double)31)];
			
		int index = 0;
		
		for(int i=0; i<tset.length; i++)
		{
			index = (tset[i]-1)/31;			
			ret[index] = ret[index] ^ (int)Math.pow(2, ((tset[i]-1) % 31));
		}
		
		return ret;
	}
	
	//#########################################################################
	// business Methods
	//#########################################################################
	
	// AND Operation -> intersection of two Transactcionlists.
	public TransactionList AND(TransactionList tl)
	{	
		TransactionList retV = new TransactionList(tl.list.length * 31);
		
		for(int i=0; i<tl.list.length; i++)
		{
			retV.list[i] = this.list[i] & tl.list[i];		
		}

		return retV;
	}
	
	public static void AND(TransactionList tlo, TransactionList tl)
	{	
		for(int i=0; i<tl.list.length; i++)
		{
			tlo.list[i] = tlo.list[i] & tl.list[i];		
		}
	}
	
	public int getFirst()
	{
		int trans;
		
		for(int i=0; i<list.length; i++)
		{
			trans = 31*i;
			
			int b = list[i];
			
			if(b != 0)
			{				
				return (trans + Integer.numberOfTrailingZeros(b) + 1);
			}
		}
		return -1;
	}
	
	public ArrayList<Integer> getTransactions()
	{
		ArrayList<Integer> ret_L = new ArrayList<>();
		
		int trans;
		
		for(int i=0; i<list.length; i++)
		{
			trans = 31*i;
			
			int b = list[i];
			int c = 0;
			
			while(b != 0)
			{
				c = b & (0 - b);
				
				ret_L.add(trans + Integer.numberOfTrailingZeros(c) + 1);
				
				b = b ^ c;
			}
			
		}		
		return ret_L;
	}
	
	public void add(int i)
	{
		int index 	= (i-1)/31;			
		list[index] 	= list[index] ^ (int)Math.pow(2, ((i-1) % 31));
	}
	
	public boolean contains(TransactionList tl)
	{
		for(int i=0; i<tl.list.length; i++)
		{
			if(tl.list[i] != (this.list[i] & tl.list[i]))
				return false;
		}
		return true;
	}
	
//	public boolean contains(TransactionList tl)
//	{
//		TransactionList.AND(this,tl);
//		
//		return tl.equals(this);
//	}
	
	public int count()
	{
		int retV = 0;
		
		for(int i: this.list)
		{
			retV = retV + Integer.bitCount(i);
		}
		
		return retV;
	}
	
	public void setHashValue(int hashValue)
	{
		this.hashValue = hashValue;
	}
	
	public void assimilate(TransactionList tl)
	{
		for(int i=0; i<tl.list.length; i++)
		{
			this.list[i] = tl.list[i];
		}
	}
	
	@Override
	public boolean equals(Object obj)
	{
		TransactionList tl = (TransactionList)obj;
		
		if(tl.list.length != this.list.length) return false;
		
		for(int i=0; i<this.list.length; i++)
		{
			if(this.list[i] != tl.list[i]) return false;
		}
		
		return true;
	}
	
//	@Override
//	public int hashCode()
//	{
//		return hashValue;
//	}
	
	@Override
	public String toString()
	{
		StringBuilder sb = new StringBuilder();
		
		for(int i=0; i<this.list.length; i++)
		{
			sb.append(Integer.toBinaryString(this.list[i]) + " | ");
		}
//		sb.append(" h:" + hashCode());
		
		return sb.toString();
	}
}
